package cz.mendelu.dnaAnalyser.analyse.zdna;

import cz.mendelu.dnaAnalyser.batch.BatchService;
import cz.mendelu.dnaAnalyser.sequence.Sequence;
import cz.mendelu.dnaAnalyser.sequence.data.SequenceData;
import cz.mendelu.dnaAnalyser.sequence.data.SequenceDataRepository;
import cz.mendelu.dnaAnalyser.sequence.stream.BufferedWindow;
import cz.mendelu.dnaAnalyser.sequence.stream.Window;
import cz.mendelu.dnaAnalyser.utils.analyse.service.heatmap.Heatmap;
import cz.mendelu.dnaAnalyser.utils.exception.NotFoundException;
import cz.mendelu.dnaAnalyser.utils.repository.BaseRepository;
import cz.mendelu.dnaAnalyser.utils.service.ModelService;
import lombok.extern.slf4j.Slf4j;
import org.apache.ibatis.session.RowBounds;
import org.apache.ibatis.session.SqlSession;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Sort;

import java.util.*;

@Service
@Slf4j
public class ZdnaService extends ModelService<Zdna> {

    @Autowired
    private ZdnaRepository zdnaRepository;

    @Autowired
    private SequenceDataRepository sequenceDataRepository;

    @Autowired
    private BatchService batchService;

    @Autowired
    private ZdnaAnalyser zdnaAnalyser;

    @Autowired
    private ZdnaDataSessionService zdnaDataSessionService;

    @Override
    protected BaseRepository<Zdna, UUID> getRepository() {
        return zdnaRepository;
    }

    public Zdna executeZdnaAnalysis(Zdna zdna) {

        zdna = zdnaRepository.save(zdna);
        batchService.createBatch(zdna.getOwner(), zdna)
                .execute(uuid -> {
                    Zdna analyse = zdnaRepository
                            .findById(uuid)
                            .orElseThrow(() -> new NotFoundException(Zdna.class, uuid));
                    Sequence sequence = analyse.getSequence();
                    SequenceData sequenceData = sequenceDataRepository.load(sequence);

                    int minSequenceSize = analyse.getMinSequenceSize();
                    float threshold = analyse.getThreshold();
                    float score_gc = analyse.getScore_gc();
                    float score_gtac = analyse.getScore_gtac();
                    float score_at = analyse.getScore_at();


                    if (sequence.getLength() < minSequenceSize) {
                        throw new IllegalArgumentException(String.format("The window (%d) must be smaller than the sequence length (%d).", minSequenceSize, sequence.getLength()));
                    }

                    try (SqlSession sqlSession = zdnaDataSessionService.openDataSession(analyse)) {
                        ZdnaMapper zdnaMapper = sqlSession.getMapper(ZdnaMapper.class);

                        String sequenceDataPlain = sequenceData.toPlain();
                        Window sequenceWindow = BufferedWindow.wrap(sequenceDataPlain);

                        ArrayList<ZdnaAnalyseResult> results = zdnaAnalyser.getResults(sequenceWindow, minSequenceSize, threshold, score_gc, score_gtac, score_at);

                        results.forEach(zdnaMapper::insert);

                        analyse.setResultCount((long) results.size());
                        analyse.finishNow();
                        sqlSession.commit();
                    }
                    zdnaRepository.save(analyse);
                }).start();
        return zdna;
    }

    public Page<ZdnaAnalyseResult> returnZdnaAnalysis(UUID id, Integer sequenceStart, Integer sequenceLength, Pageable pageable) {
        Zdna analyse = findOne(id);

        int start = (sequenceStart == null) ? 0 : sequenceStart;
        int end = start + ((sequenceLength == null) ? analyse.getSequence().getLength() : sequenceLength);

        try (SqlSession sqlSession = zdnaDataSessionService.openDataSession(analyse)) {
            ZdnaMapper zdnaMapper = sqlSession.getMapper(ZdnaMapper.class);

            RowBounds rowBounds = new RowBounds((int) pageable.getOffset(), pageable.getPageSize());
            String orderName = "POSITION";
            Sort.Direction orderDirection = Sort.Direction.ASC;
            Sort.Order order = getSortOrderOrNull(pageable);
            if (order != null) {
                orderName = order.getProperty().toUpperCase();
                orderDirection = order.getDirection();
            }

            List<ZdnaAnalyseResult> list = (orderDirection == Sort.Direction.ASC)
                    ? zdnaMapper.getAllBetweenPositionOrderAsc(start, end, orderName, rowBounds)
                    : zdnaMapper.getAllBetweenPositionOrderDesc(start, end, orderName, rowBounds);
            int total = zdnaMapper.countAllBetweenPosition(start, end);

            return new PageImpl<>(list, pageable, total);
        }
    }

    public Heatmap getHeatmap(UUID id, Integer segmentsCount, Integer from, Integer to) {
        Zdna analyse = findOne(id);
        return zdnaDataSessionService.heatmap(ZdnaMapper.class, analyse, segmentsCount, from, to);
    }

    public Zdna modifyTags(UUID id, Set<String> tags) {
        Zdna analyse = zdnaRepository.findById(id)
                .orElseThrow(() -> new NotFoundException(Zdna.class, id));
        analyse.setTags(tags);
        zdnaRepository.save(analyse);
        return analyse;
    }

    private Sort.Order getSortOrderOrNull(Pageable pageable) {
        if (pageable == null) return null;
        Iterator<Sort.Order> orderIterator = pageable.getSort().iterator();
        if (!orderIterator.hasNext()) return null;
        return orderIterator.next();
    }

    public int avgZdnaLen(UUID id) {
        Zdna analyse = findOne(id);
        int result;

        try (SqlSession sqlSession = zdnaDataSessionService.openDataSession(analyse)) {
            ZdnaMapper zdnaMapper = sqlSession.getMapper(ZdnaMapper.class);
            ArrayList<ZdnaAnalyseResult> analysisZdena = zdnaMapper.getAll();

            int totalZdnaLen = analysisZdena.stream()
                    .mapToInt(e -> e.length)
                    .sum();

            result = analysisZdena.size() != 0
                ? totalZdnaLen / analysisZdena.size()
                : 0;
        }

        return result;
    }

    public int totalZdnaLen(UUID id) {
        int totalZdnaLen;

        Zdna analyse = findOne(id);

        try (SqlSession sqlSession = zdnaDataSessionService.openDataSession(analyse)) {
            ZdnaMapper zdnaMapper = sqlSession.getMapper(ZdnaMapper.class);
            ArrayList<ZdnaAnalyseResult> analysisZdena = zdnaMapper.getAll();

            totalZdnaLen = analysisZdena.stream()
                    .mapToInt(e -> e.length)
                    .sum();
        }

        return totalZdnaLen;
    }

}

