package cz.mendelu.dnaAnalyser.user;

import cz.mendelu.dnaAnalyser.jwt.JwtException;
import cz.mendelu.dnaAnalyser.sequence.Sequence;
import cz.mendelu.dnaAnalyser.sequence.SequenceRepository;
import cz.mendelu.dnaAnalyser.utils.exception.NotFoundException;
import cz.mendelu.dnaAnalyser.email.model.Email;
import cz.mendelu.dnaAnalyser.email.service.EmailService;
import lombok.extern.log4j.Log4j;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.RandomStringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.dao.DataIntegrityViolationException;
import org.springframework.jdbc.core.BeanPropertyRowMapper;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.security.core.userdetails.UserDetails;
import org.springframework.security.core.userdetails.UserDetailsService;
import org.springframework.security.core.userdetails.UsernameNotFoundException;
import org.springframework.security.crypto.bcrypt.BCryptPasswordEncoder;
import org.springframework.security.crypto.password.PasswordEncoder;
import org.springframework.stereotype.Service;

import javax.transaction.Transactional;
import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.Stream;

@Slf4j
@Service
public class UserService implements UserDetailsService {

    private final PasswordEncoder passwordEncoder = new BCryptPasswordEncoder();

    @Value("${sequence.defaults}")
    private String[] defaultsSequences;

    @Autowired
    private UserRepository userRepository;

    @Autowired
    private JdbcTemplate jdbcTemplate;

    @Autowired
    private EmailService emailService;

    @Autowired
    private SequenceRepository sequenceRepository;

    public User registrationUser(String login, String password) {
        try {
            Date now = new Date();
            User user = User.builder()
                    .username(login)
                    .password(encoderPassword(password))
                    .createdDate(now)
                    .lastPasswordResetDate(now)
                    .authorities(EnumSet.of(Authority.ROLE_USER))
                    .enabled(true)
                    .build();
            user = userRepository.save(user);
            createDefaultData(user);
            return user;
        } catch (DataIntegrityViolationException e) {
            throw new UserLoginAlreadyExistsException("User with login already exists.", e);
        }
    }

    public void createDefaultData(User user) {
        if (defaultsSequences == null)
            return;

        List<UUID> sequencesUUID = Stream.of(defaultsSequences)
                .map(UUID::fromString)
                .collect(Collectors.toList());

        List<Sequence> userSequence = new ArrayList<>();
        for (UUID uuid: sequencesUUID) {
            Optional<Sequence> result = userRepository.findDefaultSequenceById(uuid);
            if (! result.isPresent()) {
                log.warn("Default sequence with ID: {} not found", uuid);
                continue;
            }
            Sequence s = result.get();
            Sequence newUserSequence = Sequence.builder()
                    .id(UUID.randomUUID())
                    .owner(user)
                    .bufferId(s.getBufferId())
                    .name(s.getName())
                    .created(s.getCreated())
                    .type(s.getType())
                    .circular(s.getCircular())
                    .length(s.getLength())
                    .ncbi(s.getNcbi())
                    .tags(s.getTags())
                    .fastaComment(s.getFastaComment())
                    .build();
            userSequence.add(newUserSequence);
        }
        sequenceRepository.saveAll(userSequence);
    }

    public void delete(UUID id) {
        User user = userRepository
                .findById(id)
                .orElseThrow(() -> new NotFoundException(User.class, id));
        user.setEnabled(false);
        userRepository.save(user);
    }

    public User changePassword(UUID id, String oldPassword, String newPassword) {
        User user = userRepository
                .findById(id)
                .orElseThrow(() -> new NotFoundException(User.class, id));
        if (!checkPassword(user, oldPassword)) {
            throw new UserPasswordChangeException("Old password is not correct.");
        }

        user.setPassword(encoderPassword(newPassword));
        user.setLastPasswordResetDate(new Date());
        user = userRepository.save(user);
        return user;
    }

    @Transactional
    public void resetPassword(String username) {
        log.info("Reset password for {}", username);
        User user = userRepository.findByUsername(username)
                .orElseThrow(() -> {
                    log.warn("Possible attack detected: an attempt to change the password of a non-existent user '{}'.", username);
                    return new NotFoundException(String.format("No user found with username '%s'.", username));
                });

        String newPassword = RandomStringUtils.randomAlphanumeric(8);
        user.setPassword(encoderPassword(newPassword));
        user.setLastPasswordResetDate(new Date());

        Email email = new Email(username);
        email.setSubject("Password reset");
        email.getContentBuilder()
                .append("Your account: ")
                .append(username)
                .append("\r\nYour new password is: " )
                .append(newPassword);
        emailService.sendEmail(email);
        userRepository.save(user);
    }

    public boolean checkPassword(User user, String password) {
        return passwordEncoder.matches(password, user.getPassword());
    }

    public String encoderPassword(String password) {
        return passwordEncoder.encode(password);
    }

    @Override
    public UserDetails loadUserByUsername(String username) throws UsernameNotFoundException {
        Optional<User> oUser = userRepository.findByUsername(username);
        if (oUser.isPresent()) {
            return oUser.get();
        }

        return userRepository
                .findById(UUID.fromString(username))
                .orElseThrow(() -> new UsernameNotFoundException(String.format("No user found with username '%s'.", username)));
    }

    /* NA základě chyby v issue #356 jsem nahradil JPA za JdbcTemplate */
    public User findOne(UUID userId) {
        User user = jdbcTemplate.queryForObject("select * from users where id=?",
                new Object[]{userId},
                BeanPropertyRowMapper.newInstance(User.class));

        if (user == null) {
            throw new NotFoundException(User.class, userId);
        }

        List<Authority> authorities = jdbcTemplate.queryForList("select authority_id from user_autority where user_id=?",
                new Object[]{userId}, Authority.class);

        user.setAuthorities(new HashSet<>(authorities));
        return user;
    }
}
